"""Api events enumerations."""

from .. import enm

__all__ = (
    'ErrorCode',
    'ErrorMap',
    'ErrorMessage',
    *enm.__all__
    )

from .. enm import *

from . import cfg
from . import lib


class Constants(cfg.Constants):
    """Modules specific to Api events Enums."""


class ErrorCode(lib.enum.Enum):
    """Common HTTP error codes."""

    RequestError              = 400
    NotAuthenticatedError     = 401
    NotAuthorizedError        = 403
    ResourceNotFoundError     = 404
    MethodNotAllowedError     = 405
    ResourceLockedError       = 423
    RateLimitedError          = 429
    UnexpectedError           = 500
    MethodNotImplementedError = 501


class ErrorMap(lib.enum.Enum):
    """Map of python exceptions to common HTTP errors."""

    SyntaxError            = 'RequestError'
    ConnectionRefusedError = 'NotAuthenticatedError'
    PermissionError        = 'NotAuthorizedError'
    FileNotFoundError      = 'ResourceNotFoundError'
    ModuleNotFoundError    = 'MethodNotAllowedError'
    OverflowError          = 'RateLimitedError'
    Exception              = 'UnexpectedError'
    NotImplementedError    = 'MethodNotImplementedError'

    TypeValidationError    = 'RequestError'


class ErrorMessage(lib.enum.Enum):
    """Common HTTP error messages by code."""

    _400 = 'Operation could not be completed due to an error with the request.'  # noqa
    _401 = 'Must be authenticated to complete the request.'
    _403 = 'Not authorized to complete the request.'
    _404 = 'Requested resource could not be found at the specified location.'
    _405 = 'Method not allowed for the requested resource.'
    _423 = 'Requested resource is currently locked for modification.'
    _429 = 'Too many requests.'
    _500 = 'An unexpected error occurred'
    _501 = 'Method not yet implemented for the requested resource.'



"""Api events exceptions module."""

from .. import exc

__all__ = (
    'HTTPError',
    'MethodNotAllowedError',
    'MethodNotImplementedError',
    'NotAuthenticatedError',
    'NotAuthorizedError',
    'RateLimitedError',
    'RequestError',
    'ResourceNotFoundError',
    'ResourceLockedError',
    'UnexpectedError',
    *exc.__all__
    )

from ..  exc import *


class HTTPError(Exception):
    """Base HTTP Error class."""


class RequestError(HTTPError):
    """Operation could not be completed due to an error with the request."""


class NotAuthenticatedError(HTTPError):
    """Must be authenticated to complete the request."""


class NotAuthorizedError(HTTPError):
    """Not authorized to complete the request."""


class ResourceNotFoundError(HTTPError):
    """Requested resource could not be found at the specified location."""


class ResourceLockedError(HTTPError):
    """Requested resource is currently locked for modification."""


class RateLimitedError(HTTPError):
    """Too many requests."""


class MethodNotAllowedError(HTTPError):
    """Method not allowed for the requested resource."""


class MethodNotImplementedError(HTTPError):
    """Method not yet implemented for the requested resource."""


class UnexpectedError(HTTPError):
    """An unexpected error occurred."""



"""
Api Overview
================

**Author:** dan@1howardcapital.com

**Summary:** Api module.

---

Usage
-----

```sh
$ ft3 api ${PACKAGE_NAME}

```

"""

__all__ = (
    'api_from_package',
    'cfg',
    'enm',
    'exc',
    'events',
    'lib',
    'obj',
    'server',
    'typ',
    'utl',
    'Api',
    'File',
    'Handler',
    'Request',
    'Response',
    'FILES',
    'OBJECTS'
    )

from . import cfg
from . import enm
from . import exc
from . import events
from . import lib
from . import obj
from . import server
from . import typ
from . import utl

from . events import Handler, Request, Response
from . obj import Api, File, FILES, OBJECTS
from . utl import api_from_package



"""Api objects module."""

__all__ = (
    'Component',
    'Content',
    'File',
    'Header',
    'Healthz',
    'Info',
    'Operation',
    'Api',
    'Parameter',
    'Path',
    'RequestBody',
    'ResponseObject',
    'Schema',
    'SecurityScheme',
    'ServerVariable',
    'ServerObject',
    'Tag',
    'FILES',
    'OBJECTS',
    'RESPONSE_HEADERS'
    )

from .. import cli
from .. import core

from .. import Field, Object

from . import cfg
from . import enm
from . import lib
from . import typ


class Constants(cfg.Constants):
    """Constant values specific to this file."""


FILES: dict[str, 'File'] = {}
"""All `Files` served by the API."""

OBJECTS: dict[str, 'type[typ.Object]'] = {}
"""All `Objects` served by the API."""


class Component(Object):
    """
    [OpenAPI](https://swagger.io/docs/specification/components/) Component.

    """

    _ref_: Field[lib.t.Optional[typ.AnyString]] = None


class Schema(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#schema-object) Schema Object.

    """

    type_: Field[list[typ.ApiType]] = Field(
        default=[enm.Type.string.value],
        enum=enm.Type
        )
    format_: Field[lib.t.Optional[typ.ApiFormat]] = Field(
        default=None,
        enum=enm.Format
        )

    description: Field[lib.t.Optional[str]] = None

    default: Field[lib.t.Optional[typ.ApiTypeValue]] = None
    enum: Field[lib.t.Optional[typ.Enum]] = None

    required: Field[lib.t.Optional[list[typ.string[typ.camelCase]]]] = None

    min_length: Field[lib.t.Optional[int]] = None
    max_length: Field[lib.t.Optional[int]] = None
    pattern: Field[lib.t.Optional[str]] = None

    minimum: Field[lib.t.Optional[float]] = None
    exclusive_minimum: Field[lib.t.Optional[bool]] = None
    maximum: Field[lib.t.Optional[float]] = None
    exclusive_maximum: Field[lib.t.Optional[bool]] = None
    multiple_of: Field[lib.t.Optional[float]] = None

    min_items: Field[lib.t.Optional[int]] = None
    max_items: Field[lib.t.Optional[int]] = None
    unique_items: Field[lib.t.Optional[bool]] = None

    read_only: Field[lib.t.Optional[bool]] = None
    write_only: Field[lib.t.Optional[bool]] = None

    items_: Field[lib.t.Optional['Schema']] = None
    properties: Field[
        lib.t.Optional[dict[typ.string[typ.camelCase], 'Schema']]
        ] = None

    all_of: Field[lib.t.Optional[list['Schema']]] = None
    any_of: Field[lib.t.Optional[list['Schema']]] = None
    one_of: Field[lib.t.Optional[list['Schema']]] = None

    def __post_init__(self) -> None:
        if isinstance(self.enum, lib.enum.EnumMeta):  # pragma: no cover
            self.enum = sorted(self.enum._value2member_map_)
        return super().__post_init__()

    @classmethod
    def from_obj(
        cls,
        obj: type[Object],
        /,
        **kwargs: lib.t.Any
        ) -> 'Schema':
        """Parse Schema definition from `Object`."""

        oname: typ.string[typ.PascalCase] = kwargs.pop('name', obj.__name__)
        kwargs.setdefault('ref', oname)

        properties = {
            (
                fname := (
                    core
                    .strings
                    .utl
                    .snake_case_to_camel_case(field_name)
                    )
                ): cls.from_type(
                    ref=fname + oname,
                    **{
                        k: v
                        for k, v
                        in field.items()
                        if k not in Constants.SKIP_FIELDS
                        }
                    )
            for field_name, field
            in obj.__dataclass_fields__.items()
            }

        requirements = [
            (
                core
                .strings
                .utl
                .snake_case_to_camel_case(field_name)
                )
            for field_name, field
            in obj.__dataclass_fields__.items()
            if field.required
            ]

        return cls(
            properties=properties,
            required=requirements or None,
            type_=[enm.Type.object.value],
            **kwargs
            )

    @classmethod
    def from_type(
        cls,
        /,
        *,
        type_: lib.t.Any = None,
        **kwargs: lib.t.Any
        ) -> 'Schema':
        """Parse Schema definition from python `type`."""

        typ_ = kwargs.pop('type', type_)
        types_: list[typ.ApiType] = []

        if (
            typ.utl.check.is_union(typ_)
            or typ.utl.check.is_wrapper_type(typ_)
            ):
            schemae = [
                cls.from_type(type_=tp)
                for tp
                in typ.utl.check.get_type_args(typ_)
                ]
            return cls(any_of=schemae, **kwargs)
        elif typ.utl.check.is_typevar(typ_):
            if typ_.__constraints__:
                schemae = [
                    cls.from_type(type_=tp)
                    for tp
                    in typ_.__constraints__
                    ]
                return cls(one_of=schemae, **kwargs)
            elif typ_.__bound__:
                return cls(
                    all_of=[cls.from_type(type_=typ_.__bound__)],
                    **kwargs
                    )
            else:  # pragma: no cover
                return cls(**kwargs)
        elif typ.utl.check.is_literal(typ_):
            literal_tp = typ.utl.check.get_type_args(typ_)[0]
            return cls.from_type(type_=literal_tp, **kwargs)
        elif typ.utl.check.is_object_type(typ_):
            return cls.from_obj(typ_, **kwargs)
        elif typ.utl.check.is_uuid_type(typ_):
            types_.append(enm.Type.string.value)
            return cls(
                type_=types_,
                format_=enm.Format.uuid.value,
                **kwargs
                )
        elif typ.utl.check.is_typed(typ_):
            types_.append(enm.Type.object.value)
            return cls(
                properties={
                    core.strings.utl.snake_case_to_camel_case(annotation): (
                        cls.from_type(type_=tp)
                        )
                    for annotation, tp
                    in typ_.__annotations__.items()
                    },
                type_=types_,
                **kwargs
                )
        elif typ.utl.check.is_array_type(typ_):
            types_.append(enm.Type.array.value)
            if (tps := typ.utl.check.get_args(typ_)):
                tp = tps[0]
            else:
                tp = str
            return cls(
                type_=types_,
                items_=cls.from_type(type_=tp),
                **kwargs
                )
        elif typ.utl.check.is_mapping_type(typ_):
            types_.append(enm.Type.object.value)
            return cls(type_=types_, **kwargs)
        elif typ.utl.check.is_bool_type(typ_):
            types_.append(enm.Type.boolean.value)
            return cls(
                type_=types_,
                format_=enm.Format.boolean.value,
                **kwargs
                )
        elif typ.utl.check.is_none_type(typ_):
            types_.append(enm.Type.null.value)
            return cls(type_=types_, **kwargs)
        elif typ.utl.check.is_number_type(typ_):
            otp = lib.t.get_origin(typ_) or typ_
            if issubclass(otp, lib.decimal.Decimal):
                types_.append(enm.Type.number.value)
                return cls(
                    type_=types_,
                    format_=enm.Format.double.value,
                    **kwargs
                    )
            elif issubclass(otp, int):
                types_.append(enm.Type.integer.value)
                return cls(
                    type_=types_,
                    format_=enm.Format.int32.value,
                    **kwargs
                    )
            elif issubclass(otp, float):
                types_.append(enm.Type.number.value)
                return cls(
                    type_=types_,
                    format_=enm.Format.float.value,
                    **kwargs
                    )
            else:  # pragma: no cover
                types_.append(enm.Type.number.value)
                return cls(type_=types_, **kwargs)
        elif typ.utl.check.is_datetime_type(typ_):
            types_.append(enm.Type.string.value)
            return cls(
                type_=types_,
                format_=enm.Format.datetime.value,
                **kwargs
                )
        elif typ.utl.check.is_date_type(typ_):
            types_.append(enm.Type.string.value)
            return cls(
                type_=types_,
                format_=enm.Format.date.value,
                **kwargs
                )
        else:
            return cls(**kwargs)


class Parameter(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#parameter-object) Parameter Object.

    """

    name: Field[typ.string[typ.camelCase]]

    in_: Field[str] = Field(
        default=enm.ParameterLocation.query.value,
        enum=enm.ParameterLocation
        )

    description: Field[lib.t.Optional[str]] = None

    required: Field[bool] = False
    deprecated: Field[lib.t.Optional[bool]] = None

    schema: Field[lib.t.Optional[Schema]] = None


class Header(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#header-object) Header Object.

    """

    description: Field[lib.t.Optional[str]] = None
    schema: Field[lib.t.Optional[Schema]] = None


class Content(Component):
    """OpenAPI content for ContentType."""

    schema: Field[lib.t.Optional[Schema]] = None


class RequestBody(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#request-body-object) Request Body Object.

    """

    description: Field[lib.t.Optional[str]] = None
    content: Field[dict[typ.ContentType, Content]] = {
        enm.ContentType.json.value: Content()
        }


class ResponseObject(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#response-object) Request Body Object.

    """

    description: Field[str]

    headers: Field[lib.t.Optional[dict[str, Header]]] = None
    content: Field[dict[typ.ContentType, lib.t.Optional[Content]]] = {
        enm.ContentType.text.value: None
        }


class Tag(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#tag-object) Tag Object.

    """

    name: Field[str]
    description: Field[lib.t.Optional[str]] = None


class Operation(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#operation-object) Operation Object.

    """

    description: Field[lib.t.Optional[str]] = None
    summary: Field[lib.t.Optional[str]] = None

    tags: Field[lib.t.Optional[list[str]]] = None

    parameters: Field[lib.t.Optional[list[Parameter]]] = None
    request_body: Field[lib.t.Optional[RequestBody]] = None

    responses: Field[lib.t.Optional[dict[str, ResponseObject]]] = None

    @property
    def path_uri(self) -> str:
        """Path URI."""

        path_params = [
            param._ref_
            for param
            in (self.parameters or ())
            if param.in_ == enm.ParameterLocation.path.value
            ]

        uri = ''
        for tag_ in (self.tags or ()):
            for tag in tag_.split(':'):
                name = tag[0].lower() + tag[1:]
                uri += ('/' + core.strings.utl.pluralize(name))
                if tag in path_params:
                    uri += ('/{' + name + 'Id}')

        return uri


class Path(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#path-item-object) Path Item Object.

    """

    _resource_: Field[type[Object]]

    summary: Field[str]

    description: Field[lib.t.Optional[str]] = None

    delete: Field[lib.t.Optional[Operation]] = None
    get_: Field[lib.t.Optional[Operation]] = None
    options: Field[lib.t.Optional[Operation]] = None
    patch: Field[lib.t.Optional[Operation]] = None
    post: Field[lib.t.Optional[Operation]] = None
    put: Field[lib.t.Optional[Operation]] = None

    def update_options(self) -> None:
        """Update options operation for the endpoint."""

        tags: list[str] = []
        path_params: dict[str, Parameter] = {}
        for method in Constants.METHODS:
            operation: lib.t.Optional[Operation] = self[method]
            if operation is not None:
                path_params |= {
                    param._ref_: param
                    for param
                    in (operation.parameters or ())
                    if param.in_ == enm.ParameterLocation.path.value
                    and param._ref_ is not None
                    }
                for tag in (operation.tags or ()):
                    if tag not in tags:
                        tags.append(tag)

        response_obj = ResponseObject(
            description='Empty response.',
            headers=RESPONSE_HEADERS
            )

        self.options = Operation(
            summary='Options for the endpoint.',
            tags=tags,
            parameters=list(path_params.values()) or None,
            responses={'204': response_obj}
            )

        return None


class ServerVariable(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#server-variable-object) Server Variable Object.

    """

    default: Field[str]
    description: Field[lib.t.Optional[str]] = None


class ServerObject(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#server-object) Server Object.

    """

    url: Field[str]
    description: Field[lib.t.Optional[str]] = None
    variables: Field[lib.t.Optional[dict[str, ServerVariable]]] = None


class Info(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#info-object) Info Object.

    """

    title: Field[str]
    version: Field[str]

    summary: Field[lib.t.Optional[str]] = None
    description: Field[lib.t.Optional[str]] = None
    terms_of_service: Field[lib.t.Optional[str]] = None


class File(Object):
    """A file object, useful for serving static files."""

    path: Field[str]
    """The path at which to serve the file."""

    content: Field[bytes | str]
    """File content."""

    content_type: Field[str] = enm.ContentType.text.value

    def __post_init__(self) -> None:
        FILES[self.path] = self
        return super().__post_init__()


class Api(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#openapi-object) OpenAPI Object.

    """

    info: Field[Info]
    openapi: Field[str] = Constants.VERSION

    paths: Field[dict[str, Path]] = {}
    tags: Field[list[Tag]] = []

    servers: Field[lib.t.Optional[list[ServerObject]]] = None

    @classmethod
    def register(cls, obj_: type[typ.ObjectType]) -> type[typ.ObjectType]:
        """Register an `Object` to be served from the API."""

        OBJECTS[obj_.__name__] = obj_

        return obj_


class SecurityScheme(Component):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#security-scheme-object) Security Scheme Object.

    """

    type_: Field[str] = Field(
        default=enm.SecuritySchemeType.apiKey.value,
        enum=enm.SecuritySchemeType
        )

    name_: Field[lib.t.Optional[str]] = None
    in_: Field[lib.t.Optional[str]] = Field(
        default=None,
        enum=enm.ApiKeyLocation
        )

    scheme: Field[lib.t.Optional[str]] = Field(
        default=None,
        enum=enm.SecurityHTTPScheme
        )

    description: Field[lib.t.Optional[str]] = None
    content: Field[dict[typ.ContentType, Content]] = {
        enm.ContentType.json.value: Content()
        }


class Healthz(Object):
    """Default application heartbeat."""

    status: Field[str] = 'OK'
    """Application status."""


RESPONSE_HEADERS = {
    header.value: Header(
        description=(
            '\\' + value
            if (value := str(enm.HeaderValue[header.name].value)) == '*'
            else value
            ),
        schema=Schema.from_type(type_=dict[str, str])
        )
    for header
    in enm.Header
    }
"""Default response headers."""


api_parser = cli.obj.parsers.add_parser(  # type: ignore[has-type]
    'api',
    formatter_class=lib.argparse.ArgumentDefaultsHelpFormatter,
    )
api_parser.add_argument(
    'package',
    help='the name or path to the package to be served',
    )
api_parser.add_argument(
    '--port',
    '-p',
    default=Constants.DEFAULT_PORT,
    help='the port to serve on',
    dest='port',
    )
api_parser.add_argument(
    '--version',
    '-v',
    default=Constants.DEFAULT_VERSION,
    help='the version of the api',
    dest='version',
    )
api_parser.add_argument(
    '--api-path',
    default=Constants.API_PATH,
    help='the base path from which to serve the API',
    dest='api_path',
    )
api_parser.add_argument(
    '--no-heartbeat',
    action='store_false',
    help='set to disinclude /healthz endpoint',
    dest='include_heartbeat',
    )
api_parser.add_argument(
    '--include-version-prefix',
    action='store_true',
    help='set to include a version prefix (ex. /v1/healthz)',
    dest='include_version_prefix',
    )



"""Api event utility functions."""

__all__ = (
    'handle_request',
    'paths_from_api'
    )

from .. import typ

from . import cfg
from . import enm
from . import exc
from . import lib
from . import obj


class Constants(cfg.Constants):
    """Constants specific to this file."""

    ID = 'id'


PATHS: list[str] = []
"""Cached, sorted path names."""


def _uri_len(uri: str) -> tuple[int, int]:
    split_ = uri.split('/')
    count_ = len([e for e in split_ if bool(obj.Pattern.PathId.match(e))])
    return count_, len(split_)


def paths_from_api(api: obj.Api) -> list[str]:
    """Cache and return cached paths from an OpenAPI spec."""

    if not PATHS:
        if api.servers:
            server = api.servers[0]
            if server.variables is not None:
                path_root = server.url.replace(
                    '{version}',
                    server.variables['version'].default
                    )
            else:  # pragma: no cover
                path_root = ''
        else:  # pragma: no cover
            path_root = ''
        PATHS.extend(
            sorted(
                [
                    path_root + path
                    for path
                    in api.paths
                    ],
                key=_uri_len,
                reverse=True
                )
            )

    return PATHS


def handle_request(
    request: obj.Request,
    api: obj.Api
    ) -> obj.Response:
    """Route an incoming request according to an OpenAPI spec."""

    path: lib.t.Optional[obj.Path] = None
    path_names = paths_from_api(api)

    if api.servers:
        server = api.servers[0]
        if server.variables is not None:
            server_root = server.url.replace(
                '{version}',
                server.variables['version'].default
                )
            server_root += '/'
        else:  # pragma: no cover
            server_root = Constants.API_PATH
    else:  # pragma: no cover
        server_root = Constants.API_PATH

    path_pattern = ''
    for path_name in path_names:
        path_pattern = (
            '/'.join(
                (
                    Constants.PATH_ID
                    if bool(obj.Pattern.PathId.match(path_element))
                    else path_element
                    for path_element
                    in path_name.split('/')
                    )
                )
            )
        if bool(lib.re.match(path_pattern, request.path)):
            path = api.paths[
                path_name.replace(server_root, Constants.API_PATH)
                ]
            break

    path_name = request.path
    status_code: typ.HttpStatusCode
    response_body: typ.CamelDict | list[typ.CamelDict] | str | bytes

    from ... import log
    from .. import FILES

    if (file := FILES.get(path_name)) is not None:
        content_type = file.content_type
        status_code = 200
        response_body = file.content
    elif path is not None:
        method: typ.string[typ.snake_case] = request.method.lower()
        obj_ = path._resource_
        callback = obj_.__operations__.get(method)
        if callback is not None:
            try:
                operation: obj.Operation = path[method]
                if operation.parameters is not None:
                    request.parse_path_params(
                        '/'.join(
                            (
                                server_root.rstrip('/'),
                                operation.path_uri.strip('/')
                                )
                            ),
                        operation
                        )
                    request.parse_query_params(method, operation, obj_)
                if operation.request_body is not None:
                    request.parse_body(operation, obj_)
                log.info(request)
                response_obj = callback(request)
            except Exception as exception:
                last_frame = lib.traceback.format_tb(exception.__traceback__)[-1]
                is_error_raised = 'raise ' in last_frame
                is_error_from_api = api.info.title in last_frame
                if is_error_raised or is_error_from_api:
                    error = obj.Error.from_exception(exception)
                else:  # pragma: no cover
                    error = obj.Error.from_exception(exc.UnexpectedError)
                log.error(error)
                content_type = enm.ContentType.json.value
                status_code = error.error_code
                response_body = error.as_response
            else:
                if response_obj is None:
                    content_type = enm.ContentType.text.value
                    status_code = 204 if method == Constants.DELETE else 200
                    response_body = ''
                elif isinstance(response_obj, str):  # pragma: no cover
                    content_type = enm.ContentType.text.value
                    status_code = 200
                    response_body = response_obj
                elif isinstance(response_obj, list):
                    content_type = enm.ContentType.json.value
                    status_code = 200
                    response_body = [o.as_response for o in response_obj]
                else:
                    content_type = enm.ContentType.json.value
                    status_code = 201 if method == Constants.POST else 200
                    response_body = response_obj.as_response
        elif method == Constants.OPTIONS:
            content_type = enm.ContentType.text.value
            status_code = 204
            response_body = ''
        else:
            error = obj.Error.from_exception(NotImplementedError)
            content_type = enm.ContentType.json.value
            status_code = error.error_code
            response_body = error.as_response
    else:
        error = obj.Error.from_exception(FileNotFoundError)
        content_type = enm.ContentType.json.value
        status_code = error.error_code
        response_body = error.as_response

    if isinstance(response_body, (bytes, str)):
        content_length = len(response_body)
    else:
        content_length = len(lib.json.dumps(response_body))

    headers = {
        header.value: enm.HeaderValue[header.name].value
        for header
        in enm.Header
        }
    headers[enm.Header.contentLength.value] = content_length
    headers[enm.Header.contentType.value] = content_type
    headers[enm.Header.date.value] = (
        lib.datetime.datetime.now(lib.datetime.timezone.utc).isoformat()
        )

    return obj.Response(
        request_id=request.id_,
        status_code=status_code,
        headers=headers,
        body=response_body
        )



"""Api server objects."""

__all__ = (
    'Server',
    )

from .. import events

from . import lib


class Server(lib.http.server.BaseHTTPRequestHandler):  # pragma: no cover
    """A very basic HTTP server."""

    handler: events.Handler

    def log_message(*args: lib.t.Any, **kwargs: lib.t.Any) -> None:  # noqa
        pass

    def respond(self, response: events.Response) -> None:
        """Respond to requestor."""

        self.send_response(response.status_code)
        for header, value in response.headers.items():
            self.send_header(header, str(value))
        self.end_headers()
        if response.body is not None:
            serialized_response = response.serialize()
            if isinstance(serialized_response, bytes):
                self.wfile.write(serialized_response)
            else:
                self.wfile.write(serialized_response.encode())

    def get_request(self) -> events.Request:
        """Get current request."""

        parsed = lib.urllib.parse.urlparse(self.path)

        if (content := self.headers.get('Content-Length')):
            if (
                data := self.rfile.read(int(content)).decode(errors='replace')
                ):
                body = lib.json.loads(data)
            else:
                body = {}
        else:
            body = None

        return events.Request(
            url=parsed.geturl(),
            path=parsed.path,
            body=body,
            headers=dict(self.headers),
            method=self.command,
            )

    def do_DELETE(self) -> None:
        """Handle a `DELETE` request."""

        request = self.get_request()
        response = self.handler(request)
        self.respond(response)

        return None

    def do_GET(self) -> None:
        """Handle a `GET` request."""

        request = self.get_request()
        response = self.handler(request)
        self.respond(response)

        return None

    def do_OPTIONS(self) -> None:
        """Handle a `OPTIONS` request."""

        request = self.get_request()
        response = self.handler(request)
        self.respond(response)

        return None

    def do_PATCH(self) -> None:
        """Handle a `PATCH` request."""

        request = self.get_request()
        response = self.handler(request)
        self.respond(response)

        return None

    def do_POST(self) -> None:
        """Handle a `POST` request."""

        request = self.get_request()
        response = self.handler(request)
        self.respond(response)

        return None

    def do_PUT(self) -> None:
        """Handle a `PUT` request."""

        request = self.get_request()
        response = self.handler(request)
        self.respond(response)

        return None



"""Static files for api."""

__all__ = (
    'favicon',
    'swagger_template'
    )

from .. import lib

ROOT = lib.os.path.dirname(__file__)

with open(lib.os.path.join(ROOT, 'favicon.ico'), 'rb') as f:
    favicon = f.read()

with open(lib.os.path.join(ROOT, 'swagger.html'), 'r') as f:
    swagger_template = f.read()



"""Api utility functions."""

__all__ = (
    'api_from_package',
    'filter_to_unique_params',
    'operation_from_object',
    'parameters_from_object',
    'paths_from_object',
    'response_type_from_object',
    'serve'
    )

from .. import core

from .. import Object

from . import cfg
from . import enm
from . import lib
from . import obj
from . import typ

from . obj import OBJECTS


class Constants(cfg.Constants):
    """Constants specific to this file."""

    ID = 'id'

    IN = 'in'
    REQUIRED = 'required'
    SCHEMA: typ.string[lib.t.Any] = 'schema'

    ONE: lib.t.Literal['ONE'] = 'ONE'
    MANY: lib.t.Literal['MANY'] = 'MANY'
    EMPTY: lib.t.Literal['EMPTY'] = 'EMPTY'


def parameters_from_object(cls: type[Object]) -> list[obj.Parameter]:
    """Generate RESTful API `Paremeter` Objects from an `Object`."""

    parameters: list[obj.Parameter] = []
    for name, field in cls.__dataclass_fields__.items():
        if name in cls.hash_fields:
            paramater_location = enm.ParameterLocation.path.value
            required = True
        else:
            paramater_location = enm.ParameterLocation.query.value
            required = False
        if name.strip('_') == Constants.ID:
            name_: typ.string[typ.camelCase] = (
                core.strings.utl.snake_case_to_camel_case(
                    core.strings.utl.camel_case_to_snake_case(cls.__name__)
                    )
                + 'Id'
                )
        else:
            name_ = core.strings.utl.snake_case_to_camel_case(name)
        parameter = obj.Parameter(
            _ref_=cls.__name__,
            name=name_,
            description=field.description,
            in_=paramater_location,
            schema=obj.Schema.from_type(
                **{
                    k: v
                    for k, v
                    in field.items()
                    if k not in Constants.SKIP_FIELDS
                    }
                ),
            required=required or field.required
            )
        parameters.append(parameter)

    return parameters


def response_type_from_object(
    cls: type[Object],
    method: typ.ApiMethod
    ) -> typ.ApiResponseType:
    """Calculate response type from an `Object`."""

    callback = cls.__operations__.get(method)  # type: ignore[call-overload]
    if callback is None:  # pragma: no cover
        return Constants.EMPTY

    tp = callback.__annotations__.get('return')

    if typ.utl.check.is_array_of_obj_type(tp):
        return Constants.MANY
    elif typ.utl.check.is_object_type(tp):
        return Constants.ONE
    else:
        return Constants.EMPTY


def filter_to_unique_params(
    parameters: list[obj.Parameter]
    ) -> list[obj.Parameter]:
    """Filters to a unique parameter set."""

    params: list[obj.Parameter] = []
    param_refs: list[str] = []

    for parameter in parameters:
        if parameter.name not in param_refs:
            param_refs.append(parameter.name)
            params.append(parameter)

    return params


def operation_from_object(
    cls: type[Object],
    method: typ.ApiMethod,
    parent_tags: lib.t.Optional[list[str]] = None,
    parent_path_parameters: lib.t.Optional[list[obj.Parameter]] = None,
    ) -> lib.t.Optional[obj.Operation]:
    """Generate RESTful API `Operation` from an `Object`."""

    parameters = parameters_from_object(cls)

    tags: list[str] = []
    if parent_tags is not None:
        tags.extend(parent_tags)
    tags.append(core.strings.utl.snake_case_to_camel_case(cls.__name__))

    if not cls.hash_fields:
        parameters.clear()

    response_type = response_type_from_object(cls, method)

    match response_type:
        case Constants.MANY:
            response_obj = obj.ResponseObject(
                description='Success response.',
                headers=obj.RESPONSE_HEADERS,
                content={
                    enm.ContentType.json.value: obj.Content(
                        schema=obj.Schema.from_type(type_=list[cls])  # type: ignore[valid-type]
                        )
                    }
                )
        case Constants.ONE:
            response_obj = obj.ResponseObject(
                description='Success response.',
                headers=obj.RESPONSE_HEADERS,
                content={
                    enm.ContentType.json.value: obj.Content(
                        schema=obj.Schema.from_obj(cls)
                        )
                    }
                )
        case _:
            response_obj = obj.ResponseObject(
                description='Empty response.',
                headers=obj.RESPONSE_HEADERS
                )

    match method:
        case Constants.DELETE:
            operation = obj.Operation(
                summary=f'Delete one {cls.__name__}.',
                tags=tags,
                parameters=filter_to_unique_params(
                    (parent_path_parameters or [])
                    + [
                        param
                        for param
                        in parameters
                        if param.in_ == enm.ParameterLocation.path.value
                        ]
                    ) or None,
                responses={'204': response_obj}
                )
        case Constants.GET if response_type == Constants.MANY:
            operation = obj.Operation(
                summary=f'Fetch many {cls.__name__}.',
                tags=tags,
                parameters=filter_to_unique_params(
                    (parent_path_parameters or [])
                    + [
                        obj.Parameter(  # type: ignore[misc]
                            **{
                                k: v
                                for k, v
                                in parameter.items()
                                if k != Constants.REQUIRED
                                and k != Constants.IN
                                }
                            )
                        for parameter
                        in parameters
                        ]
                    ) or None,
                responses={'200': response_obj}
                )
        case Constants.GET:
            operation = obj.Operation(
                summary=f'Fetch one {cls.__name__}.',
                tags=tags,
                parameters=filter_to_unique_params(
                    (parent_path_parameters or [])
                    + [
                        parameter
                        for parameter
                        in parameters
                        if parameter.in_ == enm.ParameterLocation.path.value
                        ]
                    ) or None,
                responses={'200': response_obj}
                )
        case Constants.PATCH:
            operation = obj.Operation(
                summary=f'Update one {cls.__name__}.',
                tags=tags,
                parameters=filter_to_unique_params(
                    (parent_path_parameters or [])
                    + [
                        obj.Parameter(  # type: ignore[misc]
                            **{
                                k: v
                                for k, v
                                in parameter.items()
                                if k != Constants.REQUIRED
                                }
                            )
                        if parameter.in_ != enm.ParameterLocation.path.value
                        else parameter
                        for parameter
                        in parameters
                        ]
                    ) or None,
                responses={'200': response_obj}
                )
        case Constants.POST:
            operation = obj.Operation(
                summary=f'Create one {cls.__name__}.',
                tags=tags,
                request_body=obj.RequestBody(
                    content={
                        enm.ContentType.json.value: obj.Content(
                            schema=obj.Schema.from_obj(cls)
                            )
                        }
                    ),
                parameters=filter_to_unique_params(
                    parent_path_parameters or []
                    ) or None,
                responses={'201': response_obj}
                )
        case Constants.PUT:
            operation = obj.Operation(
                summary=f'Replace one {cls.__name__}.',
                tags=tags,
                request_body=obj.RequestBody(
                    content={
                        enm.ContentType.json.value: obj.Content(
                            schema=obj.Schema.from_obj(cls)
                            )
                        }
                    ),
                parameters=filter_to_unique_params(
                    (parent_path_parameters or [])
                    + [
                        parameter
                        for parameter
                        in parameters
                        if parameter.in_ == enm.ParameterLocation.path.value
                        ]
                    ) or None,
                responses={'200': response_obj}
                )
        case _:  # pragma: no cover
            operation = None

    return operation


def paths_from_object(
    cls: type[Object],
    parent_tags: lib.t.Optional[list[str]] = None,
    parent_path_parameters: lib.t.Optional[list[obj.Parameter]] = None,
    ) -> list[obj.Path]:
    """Generate RESTful API `Path` Objects from an `Object`."""

    paths: list[obj.Path] = []

    operations_by_uri: dict[str, dict[typ.ApiMethod, obj.Operation]] = {}
    method: typ.ApiMethod
    for method, callback in cls.__operations__.items():
        if callback is not None:
            operation = operation_from_object(
                cls,
                method,
                parent_tags,
                parent_path_parameters
                )
            if operation is not None:
                operations_by_uri.setdefault(operation.path_uri, {})
                operations_by_uri[operation.path_uri][method] = operation

    tags: list[str] = []
    path_parameter_names: list[str] = []
    path_parameters: list[obj.Parameter] = []

    for path_uri, operations in operations_by_uri.items():
        path = obj.Path(  # type: ignore[misc]
            _ref_=path_uri,
            _resource_=cls,
            summary=cls.__name__,
            description=lib.textwrap.dedent(cls.__doc__) if cls.__doc__ else None,
            **operations  # type: ignore[arg-type]
            )

        for method, operation in operations.items():
            for parameter in (operation.parameters or ()):
                if (
                    parameter.in_ == enm.ParameterLocation.path.value
                    and parameter._ref_ is not None
                    and parameter._ref_ not in path_parameter_names
                    ):
                    path_parameter_names.append(parameter._ref_)
                    path_parameters.append(parameter)
            for tag in (operation.tags or ()):
                if tag not in tags:
                    tags.append(tag)

        path.update_options()
        paths.append(path)

    child_objs: list[type[Object]] = []
    for field in cls.__dataclass_fields__.values():
        if typ.utl.check.is_object_type(field.type_):
            child_objs.append(field.type_)
        elif typ.utl.check.is_array_of_obj_type(field.type_):
            tps: tuple[type[Object], ...] = (
                typ.utl.check.get_type_args(field.type_)
                )
            child_objs.append(tps[0])

    for child_obj in child_objs:
        if child_obj.hash_fields and child_obj.__name__ not in OBJECTS:
            paths.extend(
                paths_from_object(child_obj, tags, path_parameters or None)
                )

    return paths


def api_from_package(
    name: str,
    version: str,
    api_path: str,
    include_heartbeat: bool = True,
    include_version_prefix: bool = False
    ) -> obj.Api:
    """Generate a RESTful API from passed python package name."""

    package = lib.importlib.import_module(name)

    if include_heartbeat:
        from . import Request

        obj.Api.register(obj.Healthz)

        @obj.Healthz.GET
        def respond(request: Request) -> obj.Healthz:
            """Application status check."""

            return obj.Healthz()  # pragma: no cover

    if not name.startswith('.'.join((Constants.PACAKGE, 'template'))):  # pragma: no cover
        from .. template . pkg . obj import PetWithPet
        OBJECTS.pop(PetWithPet.__name__, None)

    paths: list[obj.Path] = []
    for obj_ in OBJECTS.values():
        paths.extend(paths_from_object(obj_))

    tags: list[obj.Tag] = []
    tagged: list[str] = []
    for path in paths:
        for method in Constants.METHODS:
            operation: lib.t.Optional[obj.Operation] = path[method]
            if operation is not None:
                if operation.tags is not None:
                    operation.tags = [':'.join(operation.tags)]
                    for tag in operation.tags:
                        if tag not in tagged:
                            _, _, obj_tag = tag.rpartition(':')
                            pascal_tag = obj_tag[0].upper() + obj_tag[1:]
                            if (obj_ := OBJECTS.get(pascal_tag)):
                                tagged.append(tag)
                                tags.append(
                                    obj.Tag(
                                        name=tag,
                                        description=(
                                            lib.textwrap.dedent(obj_.__doc__)
                                            if obj_.__doc__
                                            else None
                                            )
                                        )
                                    )

    info = obj.Info(
        title=name,
        version=version,
        summary='API created with ft3.',
        description=(
            lib.textwrap.dedent(package.__doc__)
            if package.__doc__
            else None
            )
        )


    if include_version_prefix:
        server = obj.ServerObject(
            url='/'.join((api_path.strip('/'), '{version}')),
            variables={'version': obj.ServerVariable(default=version)}
            )
    else:  # pragma: no cover
        server = obj.ServerObject(url=api_path)

    api = obj.Api(
        info=info,
        paths={path.pop('ref'): path for path in paths},
        tags=tags,
        servers=[server]
        )

    from . import static

    path_root = '/'.join((api_path.strip('/'), version))

    obj.File(
        path=path_root if include_version_prefix else api_path,
        content=(
            lib
            .string
            .Template(static.swagger_template)
            .safe_substitute(
                {
                    'TITLE': name,
                    'PATH': '/'.join((path_root, 'openapi.json'))
                    }
                )
            ),
        content_type=enm.ContentType.html.value
        )

    api_as_dict: dict[str, typ.AnyDict] = api.to_dict(
        camel_case=True,
        include_null=False,
        include_private=False,
        include_write_only=False,
        include_read_only=True
        )
    path_dict: typ.AnyDict
    param_dict: typ.AnyDict
    operation_dict: typ.AnyDict
    for path_dict in api_as_dict['paths'].values():
        for method in Constants.METHODS:
            params_list: list[typ.AnyDict] = []
            if (operation_dict := path_dict.get(method)):
                if 'parameters' not in operation_dict:
                    continue
                for param_dict in operation_dict['parameters']:
                    if param_dict['in'] == enm.ParameterLocation.path.value:
                        del param_dict[Constants.SCHEMA]
                    params_list.append(param_dict)
                operation_dict['parameters'] = params_list

    obj.File(
        path='/'.join((path_root, 'openapi.json')),
        content=lib.json.dumps(
            api_as_dict,
            indent=Constants.INDENT,
            default=repr
            ),
        content_type=enm.ContentType.json.value
        )
    obj.File(
        path='/favicon.ico',
        content=static.favicon,
        content_type=enm.ContentType.icon.value
        )

    return api


def serve(
    package: str,
    port: int,
    version: str,
    api_path: str,
    include_heartbeat: bool,
    include_version_prefix: bool
    ) -> None:  # pragma: no cover
    """
    CLI entrypoint for serving an application.

    ---

    Specify the name of your API package to serve it via simple http \
    server.

    ### Example

    `$ ft3 api ft3.template`

    ---

    You may specify a port as a positional argument following \
    the package name. By default, your application will be served \
    on port 80, accessible at http://localhost in your browser.

    ---

    ### DISCLAIMER

    `$ ft3 api serve` is highly insecure and should NOT be used \
    in any production environment.

    This _may_ change in the future.

    """

    from .. import log
    from . events import Handler
    from . server import Server

    Server.handler = Handler(
        api=api_from_package(
            package,
            version,
            api_path,
            include_heartbeat,
            include_version_prefix
            )
        )

    lib.socketserver.ThreadingTCPServer.allow_reuse_address = True
    lib.socketserver.ThreadingTCPServer.block_on_close = False
    lib.socketserver.ThreadingTCPServer.daemon_threads = True

    with lib.socketserver.ThreadingTCPServer(('', port), Server) as httpd:
        try:
            log.info(f'SERVING PORT: {port} (Press CTRL+C to quit)')
            httpd.serve_forever()
        except KeyboardInterrupt:
            log.warning('Keyboard interrupt received, exiting.')

    return None


obj.api_parser.set_defaults(func=serve)  # type: ignore[has-type]



"""Template objects."""

__all__ = (
    'Pet',
    'PetWithPet'
    )

from ... import Api, Field, Object

from . import enm
from . import lib


class Human(Object):
    """
    A human, may or may not be owned by a pet.

    ---

    To serve through the API as a resource at its own path, use \
    `@Api.register`.

    ### Example

    ```
    @Api.register
    class Human(Object):
        ...

    ```

    ---

    For now, since `Humans` belong to `PetWithPets`, but do \
    not have any unique identifiers that would cause them to become \
    a sub resource (i.e. like `Pet` below), they will not be surfaced \
    in the API.

    """

    name: Field[str] = 'Dale'
    """The human's name."""


class Pet(Object):
    """Pet of a pet."""

    pet_with_pet_id: Field[str]
    """The unique identifier of the parent pet."""

    id_: Field[str] = Field(
        default=lambda: lib.uuid.uuid4().hex,  # type: ignore[arg-type]
        read_only=True
        )
    """
    Unique identifier for the [sub] pet.

    ---

    Fields called 'id' (including underscore variations) are special, \
    they will automatically be combined with their class name to form \
    a clear object id.

    In thise case: `petId`.

    """

    name: Field[str]
    """Name of the [sub] pet."""

    type_: Field[str] = Field(
        default=enm.PetType.dog.value,
        required=True,
        enum=enm.PetType
        )
    """Type of [sub] pet. Allowed values: `'Cat', 'Dog'`."""

    in_: Field[str] = {
        'default': enm.PetLocation.outside.value,
        'required': False,
        'enum': enm.PetLocation
        }
    """Locations in which the [sub] pet can be."""

    is_tail_wagging: Field[bool] = True
    """Whether the [sub] pet's tail is currently wagging."""


@Api.register
class PetWithPet(Object):
    """
    A simple pet that makes for a good example.

    * This python docstring will be parsed as the default description \
    for its corresponding RESTful resource.

    * It will always be parsed as the description / docstring for the \
    Pet object itself in all auto-generated documentation.

    * It will be rendered in [markdown](https://www.markdownguide.org/cheat-sheet/).

    """

    id_: Field[str] = Field(
        default=lambda: lib.uuid.uuid4().hex,  # type: ignore[arg-type]
        read_only=True
        )
    """
    Unique identifier for the pet.

    ---

    Fields called 'id' (including underscore variations) are special, \
    they will automatically be combined with their class name to form \
    a clear object id.

    In this case: `petWithPetId`.

    """

    name: Field[str]
    """Name of the pet."""

    type_: Field[str] = Field(
        default=enm.PetType.dog.value,
        required=True,
        enum=enm.PetType
        )
    """Type of pet. Allowed values: `'Cat', 'Dog'`."""

    in_: Field[str] = Field(
        default=enm.PetLocation.outside.value,
        required=False,
        enum=enm.PetLocation
        )
    """Locations in which the pet can be."""

    is_tail_wagging: Field[bool] = True
    """Whether the pet's tail is currently wagging."""

    pets: Field[list[Pet]] = (
        lambda: [
            Pet(  # type: ignore[call-arg]
                name='Fido',
                type_='dog',
                in_='timeout',
                is_tail_wagging=False
                )
            ]
        )
    """The pet(s) of a pet."""

    human: Field[Human] = Human()
    """Sometimes pets also own humans."""

    def __post_init__(self) -> None:
        """Automatically set pets' parent ids on instantiation."""

        for pet in self.pets:  # pragma: no cover
            pet.pet_with_pet_id = self.id_

        return super().__post_init__()



"""
Summary
-------

A simple python API and package that can be copy / pasted / replaced.

_This description was sourced from the docstring of the root \
level `__init__.py` file of the `ft3.template` subpackage._

"""

__all__ = (
    'api',
    'pkg'
    )

from . import api
from . import pkg



"""Api enumerations."""

from .. import core

__all__ = (
    'ApiKeyLocation',
    'Component',
    'ContentType',
    'Format',
    'Header',
    'HeaderValue',
    'ParameterLocation',
    'SecurityHTTPScheme',
    'SecuritySchemeType',
    'Type',
    *core.enm.__all__
    )

from .. core . enm import *

from . import cfg
from . import lib


class Constants(cfg.Constants):
    """Modules specific to Api Enums."""


class Component(lib.enum.Enum):
    """
    [OpenAPI](https://swagger.io/docs/specification/components/) Components Enumeration.

    ---

    #### YAML Definition

    ```yaml
    components:
        # Reusable schemas (data models)
        schemas:
            ...
        # Reusable path, query, header and cookie parameters
        parameters:
            ...
        # Security scheme definitions (see Authentication)
        securitySchemes:
            ...
        # Reusable request bodies
        requestBodies:
            ...
        # Reusable responses, such as 401 Unauthorized or 400 Bad Request
        responses:
            ...
        # Reusable response headers
        headers:
            ...
        # Reusable examples
        examples:
            ...
        # Reusable links
        links:
            ...
        # Reusable callbacks
        callbacks:
            ...

    ```

    """

    callbacks       = 'callbacks'
    examples        = 'examples'
    headers         = 'headers'
    links           = 'links'
    parameters      = 'parameters'
    requestBodies   = 'requestBodies'
    responses       = 'responses'
    schemas         = 'schemas'
    securitySchemes = 'securitySchemes'


class Format(lib.enum.Enum):
    """
    [OpenAPI](https://swagger.io/docs/specification/data-models/data-types/#format) Type Formats Enumeration.

    ---

    Maps to python types.

    """

    boolean  = 'boolean'
    byte     = 'byte'
    date     = 'date'
    datetime = 'datetime'
    double   = 'double'
    float    = 'float'
    int32    = 'int32'
    uuid     = 'uuid'


class Type(lib.enum.Enum):
    """
    [OpenAPI](https://swagger.io/docs/specification/data-models/data-types/) Types Enumeration.

    ---

    Maps to python types.

    """

    array   = 'array'
    boolean = 'boolean'
    integer = 'integer'
    null    = 'null'
    number  = 'number'
    object  = 'object'
    string  = 'string'


class SecurityHTTPScheme(lib.enum.Enum):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#security-scheme-object) Security HTTP Scheme Enumeration.

    """

    basic  = 'basic'
    bearer = 'bearer'


class SecuritySchemeType(lib.enum.Enum):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#security-scheme-object) Security Types Enumeration.

    """

    http          = 'http'
    apiKey        = 'apiKey'
    oauth2        = 'oauth2'
    openIdConnect = 'openIdConnect'


class ApiKeyLocation(lib.enum.Enum):
    """
    ApiKey In Enumeration.

    ---

    There are three possible locations specified by the in field:

    1. `header`
    2. `query`
    3. `cookie`

    """

    header = 'header'
    query  = 'query'
    cookie = 'cookie'


class ParameterLocation(lib.enum.Enum):
    """
    [OpenAPI](https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#parameter-object) Paramater In Enumeration.

    ---

    There are four possible parameter locations specified by the in field:

    1. `path` Used together with Path Templating, where the parameter value is actually part of the operation's URL. This does not include the host or base path of the API. For example, in /items/{itemId}, the path parameter is itemId.
    2. `query` Parameters that are appended to the URL. For example, in /items?id=###, the query parameter is id.
    3. `header` Custom headers that are expected as part of the request. Note that RFC7230 states header names are case insensitive.
    4. `cookie` Used to pass a specific cookie value to the API.

    """

    path   = 'path'
    query  = 'query'
    header = 'header'
    cookie = 'cookie'


class ContentType(lib.enum.Enum):
    """Common content types."""

    any_ = '*/*'
    html = 'text/html'
    icon = 'image/x-icon'
    json = 'application/json'
    png  = 'image/png'
    text = 'text/plain'


class Header(lib.enum.Enum):
    """Default headers."""

    accessControlAllowCredentials = 'Access-Control-Allow-Credentials'
    accessControlAllowHeaders     = 'Access-Control-Allow-Headers'
    accessControlMaxAge           = 'Access-Control-Max-Age'
    accessControlAllowMethods     = 'Access-Control-Allow-Methods'
    accessControlAllowOrigin      = 'Access-Control-Allow-Origin'
    connection                    = 'Connection'
    contentLength                 = 'Content-Length'
    contentType                   = 'Content-Type'
    date                          = 'Date'


class HeaderValue(lib.enum.Enum):
    """Default headers."""

    accessControlAllowCredentials = True
    accessControlAllowHeaders     = '*'
    accessControlMaxAge           = 86400
    accessControlAllowMethods     = '*'
    accessControlAllowOrigin      = '*'
    connection                    = 'close'
    contentLength                 = '*'
    contentType                   = ContentType.json.value
    date                          = "'%Y-%m-%dT%H:%M:%S.%f%z'"



"""Template PetWithPets endpoints."""

__all__ = (
    'delete',
    'insert',
    'read',
    'replace',
    'update',
    )

from ... api import Request

from .. import pkg


@pkg.obj.PetWithPet.DELETE
def delete(request: Request) -> None:
    """Delete a single record."""

    id_ = request.path_params['petWithPetId']

    pkg.clients.DatabaseClient.delete_one(id_)

    return None


@pkg.obj.PetWithPet.GET
def read(request: Request) -> list[pkg.obj.PetWithPet]:
    """Read many records."""

    pets: list[pkg.obj.PetWithPet] = (
        pkg.clients.DatabaseClient.find_many(request.query_params)
        )

    return pets


@pkg.obj.PetWithPet.PATCH
def update(request: Request) -> pkg.obj.PetWithPet:
    """Update a single record."""

    id_ = request.path_params['petWithPetId']

    pet: pkg.obj.PetWithPet = pkg.clients.DatabaseClient.find_one(id_)

    if pet is None:
        raise pkg.exc.ResourceNotFoundError(
            'No pet could be found with that id.'
            )

    pet |= request.query_params
    pkg.clients.DatabaseClient.update_one(pet)

    return pet


@pkg.obj.PetWithPet.POST
def insert(request: Request) -> pkg.obj.PetWithPet:
    """Insert single record."""

    if isinstance(request.body, dict):
        pet = pkg.obj.PetWithPet(request.body)  # type: ignore[arg-type, call-arg]
        pkg.clients.DatabaseClient.insert_one(pet)
    else:
        raise SyntaxError

    return pet


@pkg.obj.PetWithPet.PUT
def replace(request: Request) -> pkg.obj.PetWithPet:
    """Replace a single record."""

    if isinstance(request.body, dict):
        pet = pkg.clients.DatabaseClient.update_one(
            pkg.obj.PetWithPet(  # type: ignore[misc]
                id_=request.path_params['petWithPetId'],
                **request.body
                )
            )
    else:  # pragma: no cover
        raise SyntaxError

    return pet



"""Type hinting utility functions."""

__all__ = (
    'collect_annotations',
    'finalize_type',
    'reference_and_expand_types',
    'resolve_type',
    )

from .. import cfg
from .. import lib

from . import check

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from .. import obj
    from .. import typ


class Constants(cfg.Constants):
    """Constant values specific to this file."""

    CACHED_ANNOTATIONS: 'dict[str, typ.AnyDict]' = {}
    """Local cache for typed object annotations."""

    MAX_RECURSIONS = 4


eval_type: lib.t.Callable[
    [
        'typ.AnyOrForwardRef',
        lib.t.Any,
        lib.t.Any,
        lib.t.Optional[frozenset]
        ],
    lib.t.Any
    ] = lib.t._eval_type  # type: ignore[attr-defined]
"""
Evaluate all `ForwardRef` in the given `type`.

---

For use of globalns and localns see the docstring for `get_type_hints()`.

`recursive_guard` is used to prevent infinite recursion with a recursive
`ForwardRef`.

"""


@lib.t.overload
def parse_ref_to_typ(
    ref: lib.t.ForwardRef,
    globalns: None,
    localns: 'typ.OptionalAnyDict'
    ) -> lib.t.ForwardRef: ...
@lib.t.overload
def parse_ref_to_typ(
    ref: lib.t.ForwardRef,
    globalns: 'typ.OptionalAnyDict',
    localns: 'typ.OptionalAnyDict'
    ) -> 'typ.AnyOrForwardRef': ...
def parse_ref_to_typ(
    ref: lib.t.ForwardRef,
    globalns: 'typ.OptionalAnyDict' = None,
    localns: 'typ.OptionalAnyDict' = None
    ) -> 'typ.AnyOrForwardRef':
    """Attempt to cast `ForwardRef` to `type`."""

    try:
        tp = eval_type(
            ref,
            globalns,
            localns,
            frozenset()
            )
    except NameError:
        return ref
    else:
        return tp


def parse_str_to_ref(
    typ_as_str: str,
    is_argument: bool,
    ) -> lib.t.ForwardRef:
    """Cast `str` to `ForwardRef`."""

    return lib.t.ForwardRef(
        typ_as_str,
        is_argument=is_argument,
        is_class=True
        )


@lib.t.overload
def resolve_type(
    typ_ref_or_str: 'typ.AnyType | typ.StrOrForwardRef',
    globalns: 'typ.AnyDict',
    localns: 'typ.AnyDict',
    is_argument: bool
    ) -> 'typ.AnyType | lib.t.Any': ...
@lib.t.overload
def resolve_type(
    typ_ref_or_str: 'typ.AnyType | typ.StrOrForwardRef',
    globalns: 'typ.OptionalAnyDict',
    localns: 'typ.OptionalAnyDict',
    is_argument: bool
    ) -> 'typ.AnyType | typ.AnyOrForwardRef': ...
@lib.t.overload
def resolve_type(
    typ_ref_or_str: 'typ.StrOrForwardRef',
    globalns: 'typ.OptionalAnyDict' = None,
    localns: 'typ.OptionalAnyDict' = None,
    is_argument: bool = False
    ) -> 'typ.AnyOrForwardRef': ...
def resolve_type(
    typ_ref_or_str: 'typ.AnyType | typ.StrOrForwardRef',
    globalns: 'typ.OptionalAnyDict' = None,
    localns: 'typ.OptionalAnyDict' = None,
    is_argument: bool = False
    ) -> 'typ.AnyType | typ.AnyOrForwardRef':
    """
    Attempt to resolve `str` or `ForwardRef` to `type`.

    ---

    Recursively resolves parameterized generics.

    """

    if isinstance(typ_ref_or_str, str):
        ref = parse_str_to_ref(typ_ref_or_str, is_argument)
        return resolve_type(ref, globalns, localns, is_argument)
    elif check.is_params_type(typ_ref_or_str):
        args = check.get_type_args(typ_ref_or_str)
        for arg in args:
            resolve_type(arg, globalns, localns, True)
        return typ_ref_or_str
    elif isinstance(typ_ref_or_str, lib.t.ForwardRef):
        typ_or_ref = parse_ref_to_typ(typ_ref_or_str, globalns, localns)
        if check.is_params_type(typ_or_ref):
            return resolve_type(typ_or_ref, globalns, localns, is_argument)
        else:
            return typ_or_ref
    else:
        return typ_ref_or_str


def _collect_annotations(
    __name: str,
    __annotations: 'typ.AnyDict',
    __bases: tuple[type, ...]
    ) -> 'typ.AnyDict':
    annotations: 'typ.AnyDict' = {}
    for _base in reversed(__bases):
        for __base in reversed(_base.__mro__):
            annotations |= getattr(__base, Constants.__ANNOTATIONS__, {})
    annotations |= __annotations
    # Ensure any annotations hinted for TYPE_CHECKING removed.
    annotations.pop(Constants.__ANNOTATIONS__, None)
    annotations.pop(Constants.__DATACLASS_FIELDS__, None)
    annotations.pop(Constants.__HERITAGE__, None)
    annotations.pop(Constants.__OPERATIONS__, None)
    annotations.pop(Constants.FIELDS, None)
    annotations.pop(Constants.ENUMERATIONS, None)
    annotations.pop(Constants.HASH_FIELDS, None)
    Constants.CACHED_ANNOTATIONS[__name] = annotations
    return annotations


def collect_annotations(
    typed_obj: 'obj.SupportsAnnotations | type[obj.SupportsAnnotations]'
    ) -> 'typ.AnyDict':
    """
    Get all type annotations for `typed_obj`.

    ---

    Walks `__bases__` to collect all annotations.

    """

    obj_tp = typed_obj if isinstance(typed_obj, type) else type(typed_obj)

    if obj_tp.__name__ in Constants.CACHED_ANNOTATIONS:
        return Constants.CACHED_ANNOTATIONS[obj_tp.__name__]

    return _collect_annotations(
        obj_tp.__name__,
        getattr(obj_tp, Constants.__ANNOTATIONS__, {}),
        obj_tp.__bases__
        )


def reference_and_expand_types(tp_or_ref: lib.t.Any) -> list[lib.t.Any]:
    """
    Recursively expand all `types` (and `type` args) into flat list of \
    `types` and / or `ForwardRefs`.

    """

    from .. import obj

    all_tps: list[lib.t.Any] = []
    tps: list[tuple[lib.t.Any, bool]] = [
        (tp_, False)
        for tp_
        in check.expand_types(tp_or_ref)
        ]

    recursion_count = 0

    while tps and recursion_count < Constants.MAX_RECURSIONS:
        tp, is_arg_tp = tps.pop()
        if isinstance(tp, str):
            all_tps.append(parse_str_to_ref(tp, is_arg_tp))
        elif check.is_params_type(tp):
            arg_tps = check.get_type_args(tp)
            recursion_count += 1
            tps.extend(
                [
                    (tp_, True)
                    for tp_
                    in arg_tps
                    ]
                )
            if any(isinstance(arg_tp, str) for arg_tp in arg_tps):
                tps.append(
                    (
                        obj.ForwardPattern.sub('', str(tp)),
                        False
                        )
                    )
            else:
                all_tps.append(tp)
        else:
            all_tps.append(tp)

    return all_tps


@lib.functools.cache
def finalize_type(tp_or_ref_or_str: lib.t.Any) -> lib.t.Any:
    """Recursively resolve any remaining `ForwardRefs`."""

    from .. import obj

    if isinstance(tp_or_ref_or_str, str):
        ref_tps = [parse_str_to_ref(tp_or_ref_or_str, False)]
    elif isinstance(tp_or_ref_or_str, lib.t.ForwardRef):  # pragma: no cover
        ref_tps = [tp_or_ref_or_str]
    else:
        ref_tps = [
            tp
            for tp
            in reference_and_expand_types(tp_or_ref_or_str)
            if isinstance(tp, lib.t.ForwardRef)
            ]

    has_unresolved_refs = bool(ref_tps)

    if has_unresolved_refs:
        ref_tps = list(set(ref_tps))
        tp_as_ref = parse_str_to_ref(
            obj.ForwardPattern.sub('', str(tp_or_ref_or_str)),
            False
            )
        ref_tps.append(tp_as_ref)

        modules: dict[str, lib.types.ModuleType] = {}

        for name, module in lib.sys.modules.items():
            if isinstance(module.__dict__, dict):
                for ref_tp in ref_tps:
                    *module_info, tp_info = ref_tp.__forward_arg__.split('.')
                    if (
                        name in '.'.join(module_info)
                        or tp_info in module.__dict__
                        ):
                        modules[name] = module

        namespace = {**modules}
        for ref_tp in ref_tps:
            parse_ref_to_typ(ref_tp, namespace, {})
            for name, module in modules.items():
                tp = parse_ref_to_typ(ref_tp, namespace, module.__dict__)
                if not isinstance(tp, lib.t.ForwardRef):
                    break

        return ref_tps[-1].__forward_value__ or tp_or_ref_or_str
    else:
        return tp_or_ref_or_str



"""Template Pets endpoints."""

__all__ = (
    'delete',
    'insert',
    'read',
    'replace',
    'update',
    )

from ... api import Request

from .. import pkg


@pkg.obj.Pet.DELETE
def delete(request: Request) -> None:
    """Delete a single record."""

    parent_id = request.path_params['petWithPetId']
    id_ = request.path_params['petId']

    pet_with_pet: pkg.obj.PetWithPet = (
        pkg.clients.DatabaseClient.find_one(parent_id)
        )

    if pet_with_pet is None:
        raise FileNotFoundError  # pragma: no cover

    pet_with_pet.pets = [
        pet
        for pet
        in pet_with_pet.pets
        if pet.id_ != id_
        ]

    pkg.clients.DatabaseClient.update_one(pet_with_pet)

    return None


@pkg.obj.Pet.GET
def read(request: Request) -> pkg.obj.Pet:
    """Read a single record."""

    parent_id = request.path_params['petWithPetId']
    id_ = request.path_params['petId']

    pet_with_pet: pkg.obj.PetWithPet = (
        pkg.clients.DatabaseClient.find_one(parent_id)
        )

    if pet_with_pet is None:  # pragma: no cover
        raise FileNotFoundError

    for pet in pet_with_pet.pets:
        if pet.id_ == id_:  # pragma: no cover
            return pet

    raise FileNotFoundError('No pet could be found for that `petId`.')


@pkg.obj.Pet.PATCH
def update(request: Request) -> pkg.obj.Pet:
    """Update a single record."""

    parent_id = request.path_params['petWithPetId']
    id_ = request.path_params['petId']

    pet_with_pet: pkg.obj.PetWithPet = (
        pkg.clients.DatabaseClient.find_one(parent_id)
        )

    if pet_with_pet is None:  # pragma: no cover
        raise FileNotFoundError

    for pet in pet_with_pet.pets:
        if pet.id_ == id_:
            pet |= request.query_params
            pkg.clients.DatabaseClient.update_one(pet_with_pet)
            return pet

    raise FileNotFoundError  # pragma: no cover


@pkg.obj.Pet.POST
def insert(request: Request) -> pkg.obj.Pet:
    """Insert single record."""

    if isinstance(request.body, dict):
        parent_id = request.path_params['petWithPetId']

        pet_with_pet: pkg.obj.PetWithPet = (
            pkg.clients.DatabaseClient.find_one(parent_id)
            )

        if pet_with_pet is None:
            raise FileNotFoundError  # pragma: no cover

        pet = pkg.obj.Pet(pet_with_pet_id=parent_id, **request.body)  # type: ignore[misc]
        pet_with_pet.pets.append(pet)
        pkg.clients.DatabaseClient.update_one(pet_with_pet)

        return pet
    else:
        raise SyntaxError  # pragma: no cover


@pkg.obj.Pet.PUT
def replace(request: Request) -> pkg.obj.Pet:
    """Replace a single record."""

    if isinstance(request.body, dict):
        parent_id = request.path_params['petWithPetId']
        id_ = request.path_params['petId']

        pet_with_pet: pkg.obj.PetWithPet = (
            pkg.clients.DatabaseClient.find_one(parent_id)
            )

        if pet_with_pet is None:
            raise FileNotFoundError  # pragma: no cover

        for pet in pet_with_pet.pets:
            if pet.id_ == id_:
                if pet.name == 'Sophie':
                    raise pkg.exc.CustomExampleError
                pet |= request.body
                pkg.clients.DatabaseClient.update_one(pet_with_pet)
                return pet

        raise FileNotFoundError  # pragma: no cover
    else:  # pragma: no cover
        raise SyntaxError



"""A module containing an example database integration."""

__all__ = (
    'DatabaseClient',
    )

from ... import Object

from . import lib
from . import typ


class DatabaseClient:
    """A simple, example database client."""

    DATA: dict[str, Object] = {}

    @classmethod
    def delete_one(cls, _id: str) -> lib.t.Optional[lib.Never]:
        """Delete an existing record from database."""

        if _id in cls.DATA:
            del cls.DATA[_id]
        else:  # pragma: no cover
            raise FileNotFoundError

        return None

    @classmethod
    def find_many(cls, query: dict[str, lib.t.Any]) -> list[Object]:
        """Get records from database that match the query."""

        return [
            record
            for record
            in cls.DATA.values()
            if all(record[k] == v for k, v in query.items())
            ]

    @classmethod
    def find_one(cls, _id: str) -> lib.t.Optional[Object]:
        """Get one record from database by primary key."""

        return cls.DATA.get(_id)

    @classmethod
    def insert_one(cls, record: typ.ObjectType) -> typ.ObjectType:
        """Add record to database."""

        cls.DATA[record[record.hash_fields[0]]] = record
        return record

    @classmethod
    def update_one(cls, record: typ.ObjectType) -> typ.ObjectType:
        """Update existing record in database."""

        cls.DATA[record[record.hash_fields[0]]] = record
        return record



"""Template enumerations."""

__all__ = (
    'PetLocation',
    'PetType',
    )

from . import lib


class PetLocation(lib.enum.Enum):
    """Valid Pet locations."""

    inside  = 'inside'
    outside = 'outside'
    timeout = 'timeout'


class PetType(lib.enum.Enum):
    """Valid Pet types."""

    cat = 'cat'
    dog = 'dog'



"""Api events constants."""

__all__ = (
    'Constants',
    )

from .. import cfg


class Constants(cfg.Constants):
    """Constant values specific to api events modules."""

    PATH_ID = '[a-zA-Z0-9]{1,256}'



"""Api constants."""

__all__ = (
    'Constants',
    )

from .. import core


class Constants(core.cfg.Constants):
    """Constant values specific to Api modules."""

    DEFAULT_VERSION = 'v1'
    VERSION = '3.1.0'
    API_PATH = '/'
    DEFAULT_PORT = 80
    METHODS = (
        'delete',
        'get',
        'options',
        'patch',
        'post',
        'put',
        )
    SKIP_FIELDS = (
        'required',
        )



"""Api imports."""

from .. import core

__all__ = (
    'datetime',
    'enum',
    'http',
    'importlib',
    'json',
    're',
    'socketserver',
    'string',
    't',
    'urllib',
    'uuid',
    'Never',
    'Self',
    *core.lib.__all__
    )

import http.server
import importlib
import socketserver
import string

from .. core . lib import *



"""Objects module."""

from .. import obj

__all__ = (
    'Error',
    'Handler',
    'Pattern',
    'Request',
    'Response',
    *obj.__all__
    )

from ... import core

from ... import log, Field, Object

from .. import typ

from .. obj import *

from . import cfg
from . import enm
from . import lib


class Constants(cfg.Constants):
    """Constant values specific to this file."""


class Pattern:
    """Compiled regex patterns."""

    PathId = lib.re.compile('({' + Constants.PATH_ID + '})')
    """Matches resource ids within a path uri."""


class Error(Object):
    """A stimple error message object."""

    error_message: Field[str]
    error_code: Field[typ.HttpErrorCode]

    @classmethod
    def from_exception(
        cls,
        exception: typ.ExceptionType | type[typ.ExceptionType]
        ) -> lib.Self:
        """Populate error object from an exception."""

        name_ = exception.__class__.__name__
        exc_tp: type[typ.ExceptionType] = exception.__class__

        msg: lib.t.Optional[str]
        if isinstance(exception.args, tuple) and exception.args:
            msg = str(exception.args[0])
        else:
            msg = None

        error_code: typ.HttpErrorCode = 500
        while issubclass(exc_tp, Exception):
            if name_ in enm.ErrorCode._member_names_:
                error_code = enm.ErrorCode[name_].value
                break
            elif name_ in enm.ErrorMap._member_names_:
                error_code = enm.ErrorCode[enm.ErrorMap[name_].value].value
                break
            else:
                name_ = exc_tp.__name__
            exc_tp = exc_tp.__class__

        if msg is None:
            msg = enm.ErrorMessage['_' + str(error_code)].value

        return cls(
            error_message=msg,
            error_code=error_code
            )


class Request(Object):
    """A simple request object."""

    id_: Field[str] = lambda: lib.uuid.uuid4().hex

    url: Field[str]
    path: Field[str]
    method: Field[str]

    headers: Field[dict[str, str]] = {}

    body: Field[lib.t.Any] = None

    path_params: Field[dict[typ.AnyString, str]] = {}
    query_params: Field[dict[typ.AnyString, lib.t.Any]] = {}

    def parse_body(
        self,
        operation: 'Operation',
        obj_: lib.t.Optional[type[typ.Object]] = None,
        ) -> lib.t.Optional[lib.Never]:
        """
        Parse JSON body from url string and optionally an `Object`.

        ---

        Automatically handles translation and injection of `id` params \
        for `PUT` requests.

        """

        deserialized: (
            dict[typ.string[lib.t.Any], lib.t.Any]
            | list[dict[typ.string[lib.t.Any], lib.t.Any]]
            | str | bool | int | float | typ.NoneType  # type: ignore[valid-type]
            )
        if operation.request_body is not None:
            content = operation.request_body.content.get(
                enm.ContentType.json.value
                )
        else:  # pragma: no cover
            content = None

        if (
            isinstance(self.body, str)
            and operation.request_body is not None
            and content is not None
            and (
                self.method == Constants.POST
                or self.method == Constants.PUT
                )
            ):
            str_body: str = self.body
            deserialized = lib.json.loads(str_body)
        elif operation.request_body is not None and content is not None:
            deserialized = self.body
        else:  # pragma: no cover
            deserialized = None

        id_params: dict[typ.string[typ.camelCase], str] = {}
        if (
            self.method == Constants.PUT
            and self.path_params
            and content is not None
            and content.schema is not None
            and content.schema.properties is not None
            ):
            ref_map: typ.AnyDict = {
                name_: schema._ref_
                for name_, schema
                in content.schema.properties.items()
                if schema._ref_ is not None
                }
            for path_param, value in self.path_params.items():
                if (id_name := ref_map.get(path_param)):
                    id_params[id_name] = value

        if isinstance(deserialized, dict) and obj_ is not None:
            body = {
                k: obj_.__dataclass_fields__[cname].parse(v)
                for k, v
                in deserialized.items()
                if core.strings.utl.isCamelCaseString(k)
                and isinstance(content, Content)
                and content.schema is not None
                and content.schema.properties is not None
                and k in content.schema.properties
                and (
                    cname := core.strings.utl.cname_for(k, obj_.fields)
                    ) is not None
                }
            body.update(id_params)
            self.body = body
        elif isinstance(deserialized, list) and obj_ is not None:
            body = [
                {
                    **id_params,
                    **{
                        k: obj_.__dataclass_fields__[cname].parse(v)
                        for k, v
                        in d.items()
                        if core.strings.utl.isCamelCaseString(k)
                        and isinstance(content, Content)
                        and content.schema is not None
                        and content.schema.properties is not None
                        and k in content.schema.properties
                        and (
                            cname := core.strings.utl.cname_for(k, obj_.fields)
                            ) is not None
                        }
                    }
                for d
                in deserialized
                ]
            self.body = body

        return None

    def parse_query_params(
        self,
        method: typ.string[typ.snake_case],
        operation: 'Operation',
        obj_: lib.t.Optional[type[typ.Object]] = None,
        ) -> None:
        """
        Parse query parameters from url string and optionally an \
        `Object`.

        """

        parsed = lib.urllib.parse.urlparse(self.url)
        self.query_params = {
            k: lib.urllib.parse.unquote(s[1])
            for _s
            in parsed.query.split('&')
            if (s := _s.split('='))
            and len(s) == 2
            and (
                (k := lib.urllib.parse.unquote(s[0]))
                and core.strings.utl.isCamelCaseString(k)
                )
            }
        if obj_ is not None:
            self.query_params = {
                param.name: field.parse(query_param)
                for param
                in (operation.parameters or ())
                if param.name
                and (
                    query_param := (
                        self.query_params.get(
                            param.name,
                            Constants.UNDEFINED
                            )
                        )
                    ) != Constants.UNDEFINED
                and param.schema is not None
                and (
                    cname := core.strings.utl.cname_for(
                        param.name,
                        obj_.fields
                        )
                    ) is not None
                and param.in_ == enm.ParameterLocation.query.value
                and (
                    method == Constants.PATCH
                    or not param.schema['write_only']
                    )
                and (
                    method == Constants.GET
                    or not param.schema['read_only']
                    )
                and (field := obj_.__dataclass_fields__.get(cname))
                }

        return None

    def parse_path_params(self, uri: str, operation: 'Operation') -> None:
        """Parse path parameters from a matched path uri."""

        parsed = lib.urllib.parse.urlparse(self.url)
        path_components = uri.strip('/').split('/')
        self.path_params = {
            s: lib.urllib.parse.unquote(v)
            for i, v
            in enumerate(parsed.path.strip('/').split('/'))
            if bool(Pattern.PathId.match(k := path_components[i]))
            and (
                (s := lib.urllib.parse.unquote(k[1:-1]))
                and core.strings.utl.isCamelCaseString(s)
                )
            }
        self.path_params = {
            param.name: path_param
            for param
            in (operation.parameters or ())
            if (
                path_param := (
                    self.path_params.get(
                        param.name,
                        Constants.UNDEFINED
                        )
                    )
                ) != Constants.UNDEFINED
            and param.schema is not None
            and param.in_ == enm.ParameterLocation.path.value
            }

        return None


class Response(Object):
    """A simple response object."""

    request_id: Field[str]

    status_code: Field[typ.HttpStatusCode] = 200

    headers: Field[dict[str, str]] = {}

    body: Field[lib.t.Any]

    def serialize(self) -> bytes | str:
        """JSON serialize body if not already a string."""

        if not isinstance(self.body, (bytes, str)):
            return lib.json.dumps(self.body)
        else:
            return self.body


class Handler(Object):
    """A simple request handler."""

    api: Field[Api]
    file_paths: Field[list[str]] = []

    def __call__(self, request: Request) -> Response:
        """Call on a request to receive a response."""

        from . import utl

        log.info(request)

        response = utl.handle_request(request, self.api)

        if response.status_code < 400:
            log.info(response)
        else:
            log.warning(response)

        return response



"""Static files for docs gen."""

__all__ = (
    'config_template',
    )

from .. import lib

ROOT = lib.os.path.dirname(__file__)

with open(lib.os.path.join(ROOT, 'conf.tpl'), 'r') as f:
    config_template = f.read()



"""Template exceptions module."""

from ... import api

__all__ = (
    'CustomExampleError',
    *api.events.exc.__all__
    )

from ... api . events . exc import *


class CustomExampleError(RequestError):
    """Cannot redefine a pet called `Bark`."""



"""
Summary
-------

A simple python package that can be copy / pasted / replaced.

"""

__all__ = (
    'clients',
    'enm',
    'exc',
    'lib',
    'obj',
    'typ'
    )

from . import clients
from . import enm
from . import exc
from . import lib
from . import obj
from . import typ



"""Api typing."""

from .. import core

__all__ = (
    'ApiFormat',
    'ApiMethod',
    'ApiResponseType',
    'ApiType',
    'ApiTypeValue',
    'ContentType',
    'HttpErrorCode',
    'HttpStatusCode',
    *core.typ.__all__
    )

from .. core . typ import *

from . import lib

ApiFormat: lib.t.TypeAlias = (
    lib.t.Literal['boolean']
    | lib.t.Literal['byte']
    | lib.t.Literal['date']
    | lib.t.Literal['datetime']
    | lib.t.Literal['double']
    | lib.t.Literal['float']
    | lib.t.Literal['int32']
    | lib.t.Literal['uuid']
    )
ApiMethod: lib.t.TypeAlias = (
    lib.t.Literal['delete']
    | lib.t.Literal['get']
    | lib.t.Literal['options']
    | lib.t.Literal['patch']
    | lib.t.Literal['post']
    | lib.t.Literal['put']
    )
ApiType: lib.t.TypeAlias = (
    lib.t.Literal['array']
    | lib.t.Literal['boolean']
    | lib.t.Literal['integer']
    | lib.t.Literal['null']
    | lib.t.Literal['number']
    | lib.t.Literal['object']
    | lib.t.Literal['string']
    )
ApiResponseType: lib.t.TypeAlias = (
    lib.t.Literal['EMPTY']
    | lib.t.Literal['MANY']
    | lib.t.Literal['ONE']
    )
ApiTypeValue: lib.t.TypeAlias = (
    list['ApiTypeValue']
    | bool
    | int
    | NoneType  # type: ignore[valid-type]
    | float
    | dict[str, 'ApiTypeValue']
    | str
    )
ContentType: lib.t.TypeAlias = (
    lib.t.Literal['*/*']
    | lib.t.Literal['text/html']
    | lib.t.Literal['image/x-icon']
    | lib.t.Literal['application/json']
    | lib.t.Literal['image/png']
    | lib.t.Literal['text/plain']
    )
HttpErrorCode: lib.t.TypeAlias = (
    lib.t.Literal[400]
    | lib.t.Literal[401]
    | lib.t.Literal[403]
    | lib.t.Literal[404]
    | lib.t.Literal[405]
    | lib.t.Literal[423]
    | lib.t.Literal[429]
    | lib.t.Literal[500]
    | lib.t.Literal[501]
    )
HttpStatusCode: lib.t.TypeAlias = (
    lib.t.Literal[200]
    | lib.t.Literal[201]
    | lib.t.Literal[204]
    | lib.t.Literal[301]
    | HttpErrorCode
    )
